"""jeig - Various eigendecomposition implementations wrapped for jax."""

__version__ = "v0.1.0"
__author__ = "Martin F. Schubert <mfschubert@gmail.com>"


from jeig.jeig import EigBackend as EigBackend
from jeig.jeig import eig as eig
from jeig.jeig import set_backend as set_backend
